/// t2c_spawn.h
/// Operations with child processes, alarms, etc.

/////////////////////////////////////////////////////////////////////////////
// Copyright 2009-2010 
// Institute for System Programming of the Russian Academy of Sciences (ISPRAS)
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// 
//    http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License. 
/////////////////////////////////////////////////////////////////////////////

#ifndef T2C_SPAWN_H_1504_INCLUDED
#define T2C_SPAWN_H_1504_INCLUDED

#include <t2c_util.h> 

#ifdef __cplusplus
extern "C" {
#endif

/////////////////////////////////////////////////////////////////////////////
// Type definitions
/////////////////////////////////////////////////////////////////////////////

/// Type of a test function.
typedef ET2CTestResult (*FT2CTestFunc)();

/// Type of a function that is executed before each test function in the same
/// process as the latter. It may perform some additional initialization or 
/// service tasks. 
/// The function returns nonzero if successful, 0 otherwise.
/// If this function reports failure, t2c_spawn() will not execute the test function
/// and it will set the result code to T2C_TR_UNINITIATED.
/// It is recommended that, in case of failure, the function prints a descriptive
/// message to the logging facility used by the tests.
/// 
/// [NB] Do not confuse this function with the "global" startup function which
/// contents are mostly taken from the <STARTUP> section.
/// That "global" function is executed in the parent process and it is done only
/// once: before the first test generated from the t2c-file executes.
typedef int (*FT2CTestStartupFunc)();

/// Type of a function that is executed after each test function in the same
/// process as the latter. It may perform some additional cleanup or 
/// service tasks. 
/// 
/// [NB] Do not confuse this function with the "global" cleanup function which
/// contents are mostly taken from the <CLEANUP> section.
/// That "global" function is executed in the parent process and it is done only
/// once: after the last test generated from the t2c-file executes.
typedef void (*FT2CTestCleanupFunc)();

/////////////////////////////////////////////////////////////////////////////
// Functions
/////////////////////////////////////////////////////////////////////////////
    
/// Spawn a new process and execute the test function there or execute it in 
/// the caller's process. 
/// 
/// If 'single_process' is nonzero, the caller's process is used, otherwise
/// a new process will be created and test function (along with startup and cleanup
/// functions) will be executed there. The caller's process will wait until this
/// process finishes somehow.
/// 
/// If 'wait_time' is nonzero and 'single_process' is 0, the caller's process will
/// wait no longer than 'wait_time' seconds. If the child process has not finished 
/// by that time, it will be killed and its result code returned as T2C_TR_UNRESOLVED.
///
// [Developer's note] 
// "TIME_EXPIRED" code is non-standard so the experts recommend not to use it. 
// Anyway, the infrastructure for this code should be present in T2C. This will
// help if this kind of test verdict becomes standardized.
// [/Developer's note]
/// 
/// Note that specifying wait time greater than 65535 may result in undefined 
/// behaviour if t2c_spawn is implemented using alarm() (see the description 
/// of alarm() in the standard ISO/IEC 9945-2:2003 Information technology -- 
/// Portable Operating System Interface (POSIX) -- Part 2: System Interfaces).
/// 
/// 'wait_time' is ignored if 'single_process' is nonzero. That is, the caller's 
/// process will execute the test function in this case, no matter how long it 
/// takes.
/// 
/// If 'startup_func' is not NULL, it is executed before the test function (in 
/// the same process as the latter). If NULL, it is ignored.
///
/// If 'cleanup_func' is not NULL, it is executed after the test function (in 
/// the same process as the latter). If NULL, it is ignored.
///
/// 'test_result' must not be NULL. 
/// '*test_result' will be set by the function in any case.
/// If the test function finishes normally (i.e. is not aborted by a signal), its 
/// return value will be returned in '*test_result'. 
/// t2c_spawn() may return other result codes there, for example, if the startup
/// function reports failure, or the test function is terminated abnormally, or if
/// the wait time for the test expires and may be in some other cases.
/// 
/// The function returns nonzero on success, 0 if it fails itself for some reason
/// or if it detects abnornal termination of the process that executes the test 
/// function (if 'single_process' is 0), or if startup function fails. 
/// If 'error_descr' is not NULL, the function may specify the description
/// of the failure in '*error_descr'. This string should be freed by the caller 
/// when no longer needed.
///
/// [NB] This function is not intended to be called from within the test functions.
int
t2c_spawn(FT2CTestFunc test_func, 
    FT2CTestStartupFunc startup_func, FT2CTestCleanupFunc cleanup_func,
    unsigned int wait_time,
    int single_process,
    ET2CTestResult* test_result,
    char** error_descr);

/////////////////////////////////////////////////////////////////////////////
#ifdef __cplusplus
}
#endif

#endif // T2C_SPAWN_H_1504_INCLUDED
