/// t2c_rcat.c
/// Requirement catalogue support - implementation.

/////////////////////////////////////////////////////////////////////////////
// Copyright 2009-2010 
// Institute for System Programming of the Russian Academy of Sciences (ISPRAS)
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// 
//    http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License. 
/////////////////////////////////////////////////////////////////////////////

#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <assert.h>

#include <smap.h>
#include <mist_string_utils.h>
#include <mist_file_utils.h>
#include <parser.h>
#include <t2c_rcat.h>

/// Default size (in bytes) of a buffer containing a path in the file system
#define T2C_PATH_BUFFER_SIZE ((PATH_MAX < 2048) ? 2048 : PATH_MAX)

///////////////////////////////////////////////////////////////////////////
// Private variables
///////////////////////////////////////////////////////////////////////////

static const char* errCwdFail = 
    "failed to obtain the path to the current directory";
static const char* errChdirFail = 
    "failed to change directory to \"%s\"";
static const char* errFReadFail = 
    "failed to read file \"%s\"";
static const char* errRcatLoadFail = 
    "failed to load requirements from \"%s\":\n    %s";    
static const char* errNoXMLDecl = 
    "XML declaration is missing or invalid";
static const char* errNoRootTag = 
    "the root tag (\"%s\") is missing";
static const char* errSectionExpected = 
    "\"%s\" section is expected";
static const char* errIdAttrExpected = 
    "\"%s\" attribute is expected and it must be the only attribute of a requirement";
static const char* errBadId = 
    "invalid requirement ID: \"%s\"";
static const char* errEmptyReq = 
    "text of a requirement must not be empty";
static const char* errBadSymbolInText = 
    "text of a requirement must not contain \'<\' or \'>\' symbols "
    "(or perhaps there is something wrong with closing \"%s\" tag)";

// Whitespace characters, etc.
static const char wspace[] = " \t\n\r";
static const size_t sz_wspace = sizeof(wspace) / sizeof(wspace[0]) - 1;

///////////////////////////////////////////////////////////////////////////
// Declaration: private methods
///////////////////////////////////////////////////////////////////////////

/// Load requirements from file with name 'fname' to catalogue 'rc'.
/// The function returns nonzero if successful, 0 otherwise.
/// If the file does not exist or cannot be opened, it is not considered 
/// a failure. But if a read error occurs, out of memory condition is encountered 
/// or there is a syntax error in the file it is considered a failure.
/// In case of a failure the function may provide additional
/// information about the failure in '*error_descr' if 'error_descr'
/// is not NULL. If it does, '*error_descr' should be freed by the caller when
/// no longer needed.
int
t2c_rcat_load(const char* fname, CReqCat* rc, char** error_descr);

///////////////////////////////////////////////////////////////////////////
// [NB] Same notes as for t2c_parse_* and t2c_match_* functions (see t2c.*,
// parser.*) apply to t2c_rcat_parse_* and t2c_rcat_match_* functions.

/// Parse the contents of a rcat-file (pointer to by parser->source)
/// and load requirements to 'rc'.
EParserResult 
t2c_rcat_parse_file(CT2CParser* parser, CReqCat* rc);

/// Parse (simplified) XML declaration that should be at the beginning of the 
/// file ("<?xml version="1.0"?>"). No data is extracted, only syntax is checked
/// and the check is quite weak now.
/// If successful, the function updates parser->beg for it to point just after
/// the matched declaration.
EParserResult 
t2c_rcat_match_decl(CT2CParser* parser);

/// Replace "&quot;", "&apos;", "&lt;", "&gt;" and "&amp;" in a given string with 
/// appropriate characters: '\"', '\'', '<', '>' and '&', respectively. 
/// Return nonzero if successful, 0 otherwise (if, although unlikely, out of memory
/// condition occurs).
int
t2c_rcat_replace_special(CMistString* ms);

///////////////////////////////////////////////////////////////////////////
// Implementation: public methods
///////////////////////////////////////////////////////////////////////////

CReqCat* 
t2c_rcat_create(const char* rcat_names[], const char* rcat_dir, char** error_descr)
{
    assert(rcat_names != NULL);
    assert(rcat_dir != NULL);
    
    if (error_descr != NULL)
    {
        *error_descr = NULL;
    }
    
    // get current directory
    static char path_buf[T2C_PATH_BUFFER_SIZE];
    char* cwd = getcwd(&path_buf[0], T2C_PATH_BUFFER_SIZE);
    if (cwd == NULL)
    {
        if (error_descr != NULL)
        {
            *error_descr = (char*)strdup(errCwdFail);
        }
        return NULL;
    }
    
    // chdir to rcat_dir
    int res = chdir(rcat_dir);
    if (res != 0)
    {
        if (error_descr != NULL)
        {
            *error_descr = (char*)malloc(strlen(errChdirFail) + strlen(rcat_dir) + 1);
            if (*error_descr != NULL)
            {
                sprintf(*error_descr, errChdirFail, rcat_dir);
            }
        }
        return NULL;
    }
    
    // now we are in 'rcat_dir' directory,
    // create the string map ("catalogue") and process the files.
    CReqCat* rc = smap_create();
    if (rc == NULL)
    {
        res = chdir(cwd);
        return NULL;
    }
    
    const char* rc_ext = DEF_RCAT_EXT;
    
    for (int i = 0; rcat_names[i] != NULL; ++i)
    {
        if (*(rcat_names[i]) != '\0') // if the name is not empty ...
        {
            CMistString* rc_name = mist_string_create(rcat_names[i]);
            if (rc_name == NULL)
            {
                t2c_rcat_destroy(rc);
                res = chdir(cwd);
                return NULL;
            }
            
            if (mist_string_append_range(rc_name, rc_ext, rc_ext + strlen(rc_ext)) != MIST_OK)
            {
                mist_string_destroy(rc_name);
                t2c_rcat_destroy(rc);
                res = chdir(cwd);
                return NULL;
            }
            
            int bOK = t2c_rcat_load(rc_name->str, rc, error_descr);
            mist_string_destroy(rc_name);
            
            if (!bOK)
            {
                t2c_rcat_destroy(rc);
                res = chdir(cwd);
                return NULL;
            }
        }
    }
    
    // return to the original working directory
    res = chdir(cwd);
    if (res != 0)
    {
        if (error_descr != NULL)
        {
            *error_descr = (char*)malloc(strlen(errChdirFail) + strlen(cwd) + 1);
            if (*error_descr != NULL)
            {
                sprintf(*error_descr, errChdirFail, cwd);
            }
        }
        t2c_rcat_destroy(rc);
        return NULL;
    }
    
    return rc;
}

size_t
t2c_rcat_get_size(const CReqCat* rc)
{
    assert(rc != NULL);
    return smap_get_size(rc);
}

const char*
t2c_rcat_lookup(CReqCat* rc, const char* id_what)
{
    assert(rc != NULL);
    assert(id_what != NULL);
    
    char* rid = (char*)strdup(id_what);
    if (rid == NULL)
    {
        return NULL;
    }
    
    const char* req = smap_lookup(rc, rid);
    free(rid);
    
    return req;
}

void
t2c_rcat_destroy(CReqCat* rc)
{
    assert(rc != NULL);
    smap_destroy(rc);
    return;
}

///////////////////////////////////////////////////////////////////////////
// Implementation: private methods
///////////////////////////////////////////////////////////////////////////

int
t2c_rcat_load(const char* fname, CReqCat* rc, char** error_descr)
{
    assert(fname != NULL);
    assert(rc != NULL);
    
    if (error_descr != NULL)
    {
        *error_descr = NULL;
    }
    
    if (!mist_file_exists(fname))
    {
        // not an error, just do nothing
        return 1;
    }
    
    char* contents = NULL;
    EMistErrorCode ec = mist_file_read_all(fname, &contents);
    if (ec == MIST_OPEN_FILE_FAILED)
    {
        // not an error, just do nothing
        return 1;
    }
    
    if (ec != MIST_OK)
    {
        // read error or/and out of memory condition
        if (error_descr != NULL)
        {
            *error_descr = (char*)malloc(strlen(errFReadFail) + strlen(fname) + 1);
            if (*error_descr != NULL)
            {
                sprintf(*error_descr, errFReadFail, fname);
            }
        }
        return 0;
    }
    assert(contents != NULL);
    
    // parse the contents of the file
    CT2CParser parser;
    t2c_parser_init(&parser, contents, contents, contents + strlen(contents));
    
    if (t2c_rcat_parse_file(&parser, rc) != PR_PARSE_OK)
    {
        if (parser.error != NULL)
        {
            char* err = (char*)malloc(strlen(errRcatLoadFail) + strlen(fname) + strlen(parser.error) + 1);
            if (err != NULL)
            {
                sprintf(err, errRcatLoadFail, fname, parser.error);
                free(parser.error);
                parser.error = err;
            }
            else
            {
                free(parser.error);
                parser.error = NULL;
            }
        }
        
        t2c_set_error(error_descr, &parser);
        free(contents);
        return 0;
    }
    
    free(contents);
    return 1;
}

EParserResult 
t2c_rcat_parse_file(CT2CParser* parser, CReqCat* rc)
{
    assert(t2c_is_parser_valid(parser));
    assert(rc != NULL);
    
    static const char* root_tag_name = "requirements";
    static const char* req_tag_name  = "req";
    static const char* rid_name      = "id";
    
    EParserResult pr;
    
    // 1. Parse XML declaration.
    pr = t2c_rcat_match_decl(parser);
    if (pr == PR_NO_MEMORY)
    {
        return pr;
    }
    
    if (pr != PR_MATCHED)
    {
        t2c_format_parse_error(parser, errNoXMLDecl, NULL);
        if (pr == PR_NOT_MATCHED)
        {
            pr = PR_SYNTAX_ERROR;
        }
        return pr;
    }
    
    // 2. Match the root tag.
    pr = t2c_match_section(parser, root_tag_name, NULL);
    if (pr == PR_NOT_MATCHED)
    {
        t2c_format_parse_error(parser, errNoRootTag, root_tag_name);
        return PR_SYNTAX_ERROR;
    }
    
    if (pr != PR_MATCHED)
    {
        return pr;
    }
    
    // 3. Parse requirements.
    parser->beg = mist_find_in_range_first_not_of(parser->beg, parser->end, wspace, sz_wspace);
    parser->old_end = parser->end; // we are to parse inside the root tag only
    
    while (parser->beg < parser->end)
    {
        CStringMap* sm = NULL;
        pr = t2c_match_section(parser, req_tag_name, &sm);
        if (pr == PR_NOT_MATCHED)
        {
            t2c_format_parse_error(parser, errSectionExpected, req_tag_name);
            return PR_SYNTAX_ERROR;
        }
        
        if (pr != PR_MATCHED)
        {
            return pr;
        }
        
        if (sm == NULL)
        {
            t2c_format_parse_error(parser, errIdAttrExpected, rid_name);
            return PR_SYNTAX_ERROR;
        }
        
        if (parser->beg == parser->end)
        {
            t2c_format_parse_error(parser, errEmptyReq, NULL);
            smap_destroy(sm);
            return PR_SYNTAX_ERROR;
        }
        
        // the contents of the section mustn't contain angle brackets
        if (mist_find_in_range_first_of(parser->beg, parser->end, "<>", 2) != parser->end)
        {
            t2c_format_parse_error(parser, errBadSymbolInText, req_tag_name);
            smap_destroy(sm);
            return PR_SYNTAX_ERROR;
        }
        
        assert(sm != NULL);
        size_t num = smap_get_size(sm);
        if (num != 1)
        {
            t2c_format_parse_error(parser, errIdAttrExpected, rid_name);
            smap_destroy(sm);
            return PR_SYNTAX_ERROR;
        }
        
        TStringPair** base = smap_as_array(sm);
        assert(base != NULL);
        assert(base[0] != NULL);
        assert(base[0]->key != NULL);
        assert(base[0]->val != NULL);
        
        if (strcmp(base[0]->key, rid_name) != 0) // unknown attribute
        {
            t2c_format_parse_error(parser, errIdAttrExpected, rid_name);
            smap_destroy(sm);
            return PR_SYNTAX_ERROR;
        }
        
        if (mist_find_first_not_of(base[0]->val, wspace, sz_wspace) == NULL)
        {
            t2c_format_parse_error(parser, errBadId, base[0]->val);
            smap_destroy(sm);
            return PR_SYNTAX_ERROR;
        }
        
        // extract the text of the requirement and make necessary replacements
        CMistString* text = mist_string_create_from_range(parser->beg, parser->end);
        if (text == NULL)
        {
            smap_destroy(sm);
            return PR_NO_MEMORY;
        }
        
        mist_string_trim(text);
        if (text->str[0] == '\0')
        {
            t2c_format_parse_error(parser, errEmptyReq, NULL);
            mist_string_destroy(text);
            smap_destroy(sm);
            return PR_SYNTAX_ERROR;
        }
        
        if (t2c_rcat_replace_special(text) == 0)
        {
            mist_string_destroy(text);
            smap_destroy(sm);
            return PR_NO_MEMORY;
        }
        
        mist_chars_to_spaces(text->str, text->str + strlen(text->str), "\t\r\n", 3);
        
        if (smap_add_element(rc, base[0]->val, text->str) == 0)
        {
            mist_string_destroy(text);
            smap_destroy(sm);
            return PR_NO_MEMORY;
        }
                
        mist_string_destroy(text);
        smap_destroy(sm);
        
        t2c_parser_to_section_end(parser);
        t2c_parser_advance(parser);
        parser->beg = mist_find_in_range_first_not_of(parser->beg, parser->end, wspace, sz_wspace);
    }
    
    return PR_PARSE_OK;
}

EParserResult 
t2c_rcat_match_decl(CT2CParser* parser)
{
    assert(t2c_is_parser_valid(parser));
    
    const char* pos = mist_find_in_range_first_not_of(parser->beg, parser->end, wspace, sz_wspace);
    if (pos == parser->end ||
        *pos != '<')
    {
        return PR_NOT_MATCHED;
    }
    
    ++pos;
    if (pos == parser->end ||
        *pos != '?')
    {
        return PR_NOT_MATCHED;
    }
    
    pos = mist_find_in_range_first_not_of(pos + 1, parser->end, wspace, sz_wspace);
    const char* end = mist_find_in_range_first_of(pos, parser->end, ">", 1);
    
    if (pos == parser->end || 
        end == parser->end || 
        *(end - 1) != '?' ||
        (end - pos) < 4)    // 4: 3 (strlen("xml")) + 1 (for '?')
    {
        return PR_NOT_MATCHED;
    }
    
    if ((pos[0] == 'X' || pos[0] == 'x') &&
        (pos[1] == 'M' || pos[1] == 'm') &&
        (pos[2] == 'L' || pos[2] == 'l'))
    { 
        parser->beg = end + 1;
        return PR_MATCHED;
    }
    else
    {
        return PR_NOT_MATCHED;
    }
}

int
t2c_rcat_replace_special(CMistString* ms)
{
    assert(ms != NULL);
    assert(ms->buf != NULL);
    assert((ms->str >= ms->buf) && (ms->str < ms->buf + ms->bsize));
    
    static const char* swhat[] = {
        "&quot;", 
        "&apos;", 
        "&lt;", 
        "&gt;", 
        "&amp;" // must be the last one because '&' is used in other sequences
    };
    
    static const char* swith[] = {
        "\"", 
        "\'", 
        "<", 
        ">", 
        "&"
    };

    assert(sizeof(swhat) == sizeof(swith));
    
    for (size_t i = 0; i < sizeof(swhat) / sizeof(swhat[0]); ++i)
    {
        EMistErrorCode ec = mist_string_replace(ms, swhat[i], swith[i]);
        if (ec != MIST_OK)
        {
            return 0;
        }
    }
    return 1;
}

/////////////////////////////////////////////////////////////////////////////
