/// t2c_util.c
/// Utility functions provided by T2C API - implementation.

/////////////////////////////////////////////////////////////////////////////
// Copyright 2009-2010 
// Institute for System Programming of the Russian Academy of Sciences (ISPRAS)
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// 
//    http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License. 
/////////////////////////////////////////////////////////////////////////////

#include <t2c_util.h>
#include <mist_file_utils.h>

#include <string.h>
#include <errno.h>
#include <assert.h>

char*
t2c_get_data_path(const char* suite_root_dir, const char* test_name, 
                  const char* rel_path)
{
    assert(suite_root_dir != NULL);
    assert(test_name != NULL);
    assert(rel_path != NULL);

    char* p1 = mist_path_sum(suite_root_dir, "testdata");
    if (p1 == NULL)
    {
        return NULL;
    }
    
    char* p2 = mist_path_sum(p1, test_name);
    free(p1);
    if (p2 == NULL)
    {
        return NULL;
    }
    
    p1 = mist_path_sum(p2, rel_path);
    free(p2);
        
    return p1;
}

void
t2c_set_result_code(ET2CTestResult* rc, ET2CTestResult new_rc)
{
    assert(rc != NULL);
    assert(*rc != T2C_TR_ERROR);
    assert(new_rc != T2C_TR_ERROR);
    
    // 1. T2C_TR_FAILED
    if (new_rc == T2C_TR_FAILED)
    {
        *rc = T2C_TR_FAILED;
    }
    
    if (*rc == T2C_TR_FAILED)
    {
        return;
    }
    
    // 2. T2C_TR_UNRESOLVED, T2C_TR_UNINITIATED
    if ((new_rc == T2C_TR_UNRESOLVED) || 
        (new_rc == T2C_TR_UNINITIATED))
    {
        if ((*rc != T2C_TR_UNRESOLVED) && 
            (*rc != T2C_TR_UNINITIATED))
        {
            *rc = new_rc;
        }
    }
    
    if ((*rc == T2C_TR_UNRESOLVED) || 
        (*rc == T2C_TR_UNINITIATED))
    {
        return;
    }
    
    // 3. T2C_TR_TIME_EXPIRED, T2C_TR_FIP
    if ((new_rc == T2C_TR_TIME_EXPIRED) || 
        (new_rc == T2C_TR_FIP))
    {
        if ((*rc != T2C_TR_TIME_EXPIRED) && 
            (*rc != T2C_TR_FIP))
        {
            *rc = new_rc;
        }
    }
    
    if ((*rc == T2C_TR_TIME_EXPIRED) || 
        (*rc == T2C_TR_FIP))
    {
        return;
    }
    
    // 4. T2C_TR_UNSUPPORTED, T2C_TR_UNTESTED
    if ((new_rc == T2C_TR_UNSUPPORTED) || 
        (new_rc == T2C_TR_UNTESTED))
    {
        if ((*rc != T2C_TR_UNSUPPORTED) && 
            (*rc != T2C_TR_UNTESTED))
        {
            *rc = new_rc;
        }
    }
    
    if ((*rc == T2C_TR_UNSUPPORTED) || 
        (*rc == T2C_TR_UNTESTED))
    {
        return;
    }
    
    // 5. T2C_TR_PASSED
    if (new_rc == T2C_TR_PASSED)
    {
        *rc = T2C_TR_PASSED;
    }
    
    if (*rc == T2C_TR_PASSED)
    {
        return;
    }
    
    assert(0 && "Bad result code");
    *rc = new_rc;
    return;
}

long
t2c_get_number(const char* str, long prev, const char* test_name,
    FT2CPrintFunc print_func)
{
    assert(str != NULL);
    assert(test_name != NULL);
    
    if (*str == '\0')
    {
        return prev;
    }
    
    char* res = NULL;
    int errno_saved = 0;
    errno = 0;
    
    long val = strtol(str, &res, 10);
    errno_saved = errno;
    errno = 0;
    assert(res != NULL);
    
    static const char* fmt = 
        "%s: unable to convert the following string to number: \"%s\"";
            
    char* descr = NULL;
    if (errno_saved != 0 || *res != '\0')
    {
        descr = (char*)malloc(strlen(fmt) + strlen(test_name) + strlen(str) + 1);
        if (descr != NULL &&
            sprintf(descr, fmt, test_name, str) >= 0)
        {
            print_func(descr);
        }
        free(descr);
        
        if (errno_saved != 0)
        {
            print_func(strerror(errno_saved));
        }
        
        return prev;
    }
    return val;
}

char* 
t2c_get_dir(const char* file_path)
{
    assert(file_path != NULL);
    assert(file_path[0] != '\0');
        
    return mist_path_get_containing_dir(file_path);
}

char*
t2c_path_sum(const char* base_path, const char* rel_path)
{
    assert(base_path != NULL);
    assert(rel_path != NULL);
    
    char* tpath = mist_path_sum(base_path, rel_path);
    if (tpath == NULL)
    {
        return NULL;
    }
    
    char* result = mist_path_absolute(tpath);
    free(tpath);
    
    return result;
}

/////////////////////////////////////////////////////////////////////////////
