#!/bin/bash

#############################################################################
# Usage:
#     ./gen_files.sh <target>
# or
#     ./gen_files.sh clean
# 
# The script invokes T2C file generator to create appropriate files for the 
# suite (both common and per-test ones) for the specified target.
# It also creates directory "t2c_api/<target>/" and copies the sources of 
# appropriate T2C API library there.
#
# For more information about file generation targets, see T2C Reference.
#
# When called with "clean" argument, the script removes all files generated 
# by T2C. The script also removes "t2c_api" directory with all its contents.
# 
#############################################################################

if [ $# -ne 1 ]; then
    printf "Usage:\n    ./gen_files.sh <target>\nor\n    ./gen_files.sh clean\n"
    exit 0
fi

# get the directory where this script resides.
WORK_DIR=$(cd `dirname $0` && pwd) 
cd ${WORK_DIR}

# subdirectories to process
SUBDIRS="${WORK_DIR}"

# root directory of T2C API libraries
API_ROOT_DIR=t2c_api

# name of the file containing the list of test groups (modules)
TEST_LIST_FILE_NAME=test_module.list

# the target to generate files for
TARGET="$1"

if [ "$1" = "clean" ]; then
	for SUITE in ${SUBDIRS}; do
		printf "Removing files generated for ${SUITE} ...\n"
		rm -f  ${suite}/Makefile
		rm -f  "${SUITE}/Makefile.am"
		rm -f  "${SUITE}/configure.ac"
		rm -f  "${SUITE}/${TEST_LIST_FILE_NAME}"
		rm -rf "${SUITE}/tests"
	done
	
	rm -rf "${API_ROOT_DIR}"
	printf "Done\n"
	exit 0;
fi

# check if t2c is available
which t2c >/dev/null 2>&1
if [ $? -ne 0 ]; then
    echo "t2c executable is not found, please make sure it is in a directory specified in PATH"
    exit 1
fi

LOG_FILE=gen_files.log
rm -f "${LOG_FILE}"

# copy API library first; do not touch API libraries for other targets
rm -rf "${API_ROOT_DIR}/${TARGET}"
mkdir -p "${API_ROOT_DIR}"

T2C_API_DIR="$(t2c --print-data-dir)/api/${TARGET}"
if [ ! -d "${T2C_API_DIR}" ]; then
	printf "T2C API directory does not exist: ${T2C_API_DIR}\n"
	exit 1
fi

cp -r ${T2C_API_DIR} ${API_ROOT_DIR}
if [ $? -ne 0 ]; then
	printf "Failed to copy T2C API library to ${API_ROOT_DIR}\n"
	exit 1
fi
    
for SUITE in ${SUBDIRS}; do
	# clear the list of test modules
	TEST_LIST_FILE="${SUITE}/${TEST_LIST_FILE_NAME}"
	printf "" > "${TEST_LIST_FILE}"

	# generate files for each .t2c-file
    T2C_FILES=$(find "${SUITE}/src" -type f -a -name '*.t2c')
    for fl in ${T2C_FILES}; do
        printf "Generating files for ${fl}... "
        t2c --target="${TARGET}" "${SUITE}" "${fl}" 2>>"${LOG_FILE}"
        if [ $? -eq 0 ]; then
            printf "Success\n"
			
			TEST_MODULE=`echo "${fl}" | sed 's#.*\/\([^\/]*\)$#\1#; s#\.t2c$##'`
			printf "${TEST_MODULE}\n" >> "${TEST_LIST_FILE}"
        else
            printf "Failed\n"
			printf "\n" >> "${LOG_FILE}"
        fi
    done
	
	# generate common files
	printf "Generating common files for ${SUITE}... "
	t2c --target=$1 --list-file="${TEST_LIST_FILE}" "${SUITE}" 2>>"${LOG_FILE}"
	if [ $? -eq 0 ]; then
		printf "Success\n"
	else
		printf "Failed\n"
	fi
	
	rm -f "${TEST_LIST_FILE}"
done

ERRORS_OCCURED=$(cat ${LOG_FILE})

if [ -n "${ERRORS_OCCURED}" ]; then
    echo "One or more errors occured, please check ${LOG_FILE} for details"
    exit 1
else
	rm -f "${LOG_FILE}"
    echo "File generation completed successfully"
fi

# Done
exit 0
