/// parser.c 
/// Utilities for parsing .t2c files.

/////////////////////////////////////////////////////////////////////////////
// Copyright 2009-2010 
// Institute for System Programming of the Russian Academy of Sciences (ISPRAS)
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// 
//    http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License. 
/////////////////////////////////////////////////////////////////////////////

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <assert.h>

#include "parser.h"

///////////////////////////////////////////////////////////////////////////
// Private variables
///////////////////////////////////////////////////////////////////////////

// Whitespace characters, etc.
static const char wspace[] = " \t\n\r";
static size_t sz_wspace = sizeof(wspace) / sizeof(wspace[0]) - 1;

static const char ws_brac[] = "> \t\n\r"; // this is used to find an end of a tag
static size_t sz_ws_brac = sizeof(ws_brac) / sizeof(ws_brac[0]) - 1;

// Format strings for various errors
static const char* errNoCloseAngleBracket = 
    "closing angle bracket ('>') is not found";
static const char* errAttrNoName = 
    "found '=' without the name of an attribute";
static const char* errAttrMissingEq = 
    "'=' is missing from the attribute specification";
static const char* errAttrMissingQuot = 
    "value of an attribute should be enclosed in quotes (\")";
static const char* errAttrDuplicate = 
    "attribute '%s' is specified more than once";
static const char* errNoClosingTag = 
    "no closing tag found for section '%s'";
static const char* errInvalidAttrName = 
    "invalid name of an attribute";
    
///////////////////////////////////////////////////////////////////////////
// Implementation: "public methods"
///////////////////////////////////////////////////////////////////////////

int
t2c_is_parser_valid(CT2CParser* parser)
{
    if (parser == NULL || parser->source == NULL || 
        parser->beg == NULL || parser->end == NULL)
    {
        return 0;
    }
    
    size_t len = strlen(parser->source);
    
    if (parser->beg < parser->source || 
        parser->beg > parser->source + len)
    {
        return 0;
    }
    
    if (parser->end < parser->beg || 
        parser->end > parser->source + len)
    {
        return 0;
    }
    
    if (parser->old_end < parser->beg || 
        parser->old_end > parser->source + len)
    {
        return 0;
    }
    
    return 1;
}

EParserResult
t2c_match_section(CT2CParser* parser, const char* name, CStringMap** attr)
{
    assert(t2c_is_parser_valid(parser));
    assert(name != NULL);
    
    if (attr != NULL)
    {
        *attr = NULL;
    }
    
    const char* beg_spare = parser->beg;
    const char* end_spare = parser->end;
    
    EParserResult pr = t2c_match_opening_tag(parser, name, attr);
    if (pr != PR_MATCHED)
    {
        // restore the state of the parser
        parser->beg = beg_spare;
        parser->end = end_spare;
        return pr;
    }
    
    pr = t2c_match_closing_tag(parser, name);
    if (pr != PR_MATCHED)
    {
        // restore the state of the parser
        parser->beg = beg_spare;
        parser->end = end_spare;
        
        if (attr != NULL && *attr != NULL)
        {
            smap_destroy(*attr);
            *attr = NULL;
        }
        
        if (pr == PR_NOT_MATCHED)
        {
            t2c_format_parse_error(parser, errNoClosingTag, name);
            return PR_SYNTAX_ERROR;
        }
        else // some other error occured
        {
            return pr;
        }
    }
    
    return PR_MATCHED;
}

EParserResult
t2c_parse_simple_section(CT2CParser* parser, const char* name, CStringMap* params)
{
    assert(t2c_is_parser_valid(parser));
    assert(name != NULL);
    assert(params != NULL);
    
    char* contents = mist_get_substring(parser->beg, parser->end);
    if (contents == NULL)
    {
        return PR_NO_MEMORY;
    }
    
    int bOK = smap_add_element(params, name, contents);
    free(contents);
    
    if (!bOK)
    {
        return PR_NO_MEMORY;
    }
    
    t2c_parser_to_section_end(parser);
    
    return PR_PARSE_OK;
}

void
t2c_set_error(char** error_descr, CT2CParser* parser)
{
    assert(t2c_is_parser_valid(parser));
        
    if (error_descr != NULL)
    {
        *error_descr = parser->error;
    }
    else if (parser->error != NULL)
    {
        // the caller does not need the description of the error,
        // so we just release it
        free(parser->error);
    }
    
    return;
}

void 
t2c_format_parse_error(CT2CParser* parser, const char* fmt, const char* arg)
{
    assert(t2c_is_parser_valid(parser));
    assert(parser->error == NULL);
    assert(fmt != NULL);
            
    static const char* fmt_all = "line %u: %s";
    size_t arg_len = (arg == NULL) ? 0 : strlen(arg);
    
    size_t len = strlen(fmt) + arg_len;
    char* buf = (char*)malloc(len + 1);
    if (buf == NULL)
    {
        return;
    }
    
    parser->error = (char*)malloc(len + MAX_NUM_DIGITS + strlen(fmt_all) + 1);
    if (parser->error != NULL)
    {   
        if (arg != NULL)
        {
            sprintf(buf, fmt, arg);
            sprintf(parser->error, fmt_all, mist_line_num_for_ptr(parser->source, parser->beg), buf);
        }
        else
        {
            sprintf(parser->error, fmt_all, mist_line_num_for_ptr(parser->source, parser->beg), fmt);
        }
    }
    
    free(buf);
    return;
}

void 
t2c_format_parse_error_no_line(CT2CParser* parser, const char* fmt, const char* arg)
{
    assert(t2c_is_parser_valid(parser));
    assert(parser->error == NULL);
    assert(fmt != NULL);
            
    size_t len = strlen(fmt) + ((arg == NULL) ? 0 : strlen(arg));

    parser->error = (char*)malloc(len + 1);
    if (parser->error != NULL)
    {   
        if (arg != NULL)
        {
            sprintf(parser->error, fmt, arg);
        }
        else
        {
            strcpy(parser->error, fmt);
        }
    }
    
    return;
}

EParserResult
t2c_match_opening_tag(CT2CParser* parser, const char* tag_name, CStringMap** attr)
{
    assert(t2c_is_parser_valid(parser));
    assert(tag_name != NULL);
    
    if (attr != NULL)
    {
        *attr = NULL;
    }
        
    const char* beg = mist_find_in_range_first_not_of(parser->beg, parser->end, wspace, sz_wspace);
    if (beg == parser->end || *beg != '<')
    {
        return PR_NOT_MATCHED;
    }
    
    // get the name of the tag
    beg = mist_find_in_range_first_not_of(beg + 1, parser->end, wspace, sz_wspace);
    if (beg == parser->end)
    {
        return PR_NOT_MATCHED;
    }
    
    const char* end = mist_find_in_range_first_of(beg, parser->end, ws_brac, sz_ws_brac);
    if (end == parser->end)
    {
        return PR_NOT_MATCHED;
    }
    
    size_t len = end - beg;
    if (len != strlen(tag_name) || 
        strncmp(beg, tag_name, len) != 0)
    {
        // name does not match
        return PR_NOT_MATCHED;
    }
    
    // get the string containing the attributes (if any)
    beg = mist_find_in_range_first_not_of(end, parser->end, wspace, sz_wspace);
    assert(beg < parser->end);
    if (*beg == '>')
    {
        // no attributes, just a bare tag
        parser->beg = beg + 1;
        return PR_MATCHED;
    }
    
    end = mist_find_in_range_first_of(beg, parser->end, ">", 1);
    if (end == parser->end)
    {
        parser->beg = beg;
        t2c_format_parse_error(parser, errNoCloseAngleBracket, NULL);
        return PR_SYNTAX_ERROR;
    }
    const char* close_br = end;
    
    end = mist_find_in_range_last_not_of(beg, end, wspace, sz_wspace);
    assert(end != close_br);
    ++end; //end now points past the end of what probably is the string of attributes
    
    // parse attributes
    CT2CParser attr_parser;
    t2c_parser_init(&attr_parser, parser->source, beg, end);
    
    EParserResult pr = t2c_parse_attributes(&attr_parser, attr);
    if (pr != PR_PARSE_OK)
    {
        t2c_set_error(&(parser->error), &attr_parser);
        return pr;
    }
    
    parser->beg = close_br + 1;
    return PR_MATCHED;
}

EParserResult
t2c_match_closing_tag(CT2CParser* parser, const char* tag_name)
{
    assert(t2c_is_parser_valid(parser));
    assert(tag_name != NULL);
    
    const char* beg = parser->beg;
    while (beg < parser->end)
    {
        beg = strstr(beg, "</");
        if (beg == NULL || beg >= parser->end)
        {
            break; // no match found
        }
        
        const char* open_brac_pos = beg;
        
        beg += 2;
        
        beg = mist_find_in_range_first_not_of(beg, parser->end, wspace, sz_wspace);
        if (beg == parser->end)
        {
            break; // no match found
        }
        
        const char* end = mist_find_in_range_first_of(beg, parser->end, ws_brac, sz_ws_brac);
        if (end == parser->end)
        {
            break; // no match found
        }
        
        size_t len = end - beg;
        if (len != strlen(tag_name) || 
            strncmp(beg, tag_name, len) != 0)
        {
            // name does not match, let's try one more time
            continue;
        }
        
        beg = mist_find_in_range_first_not_of(end, parser->end, wspace, sz_wspace);
        if (beg != parser->end && *beg == '>')
        {
            parser->end = open_brac_pos;
            return PR_MATCHED;
        }
        else
        {
            break; // no match found
        }
    }
    return PR_NOT_MATCHED;
}

EParserResult
t2c_parse_attributes(CT2CParser* parser, CStringMap** attr)
{
    assert(t2c_is_parser_valid(parser));
    if (attr == NULL)
    {
        // nothing to do
        return PR_PARSE_OK;
    }
    
    const char* old_beg = parser->beg;
    
    *attr = smap_create();
    if (*attr == NULL)
    {
        return PR_NO_MEMORY;
    }
    assert(parser->beg < parser->end);
    
    while (parser->beg < parser->end)
    {
        // extract the name of the attribute
        parser->beg = mist_find_in_range_first_not_of(parser->beg, parser->end, wspace, sz_wspace);
        if (parser->beg == parser->end)
        {
            break;
        }
        
        if (*(parser->beg) == '=')
        {
            t2c_format_parse_error(parser, errAttrNoName, NULL);
            smap_destroy(*attr);
            *attr = NULL;
            return PR_SYNTAX_ERROR;
        }
        
        const char* end = mist_find_in_range_first_of(parser->beg, parser->end, "=", 1);
        if (end == parser->end)
        {
            t2c_format_parse_error(parser, errAttrMissingEq, NULL);
            smap_destroy(*attr);
            *attr = NULL;
            return PR_SYNTAX_ERROR;
        }
                
        const char* pos = mist_find_in_range_last_not_of(parser->beg, end, wspace, sz_wspace);
        assert(pos != end);
        ++pos;
        
        char* name = mist_get_substring(parser->beg, pos);
        if (name == NULL)
        {
            smap_destroy(*attr);
            *attr = NULL;
            return PR_NO_MEMORY;
        }
        
        // check if the name is valid
        if (mist_name_is_bad(name))
        {
            t2c_format_parse_error(parser, errInvalidAttrName, NULL);
            free(name);
            smap_destroy(*attr);
            *attr = NULL;
            return PR_SYNTAX_ERROR;
        }
        
        // now extract the value of this attribute
        parser->beg = mist_find_in_range_first_not_of(end + 1, parser->end, wspace, sz_wspace);
        if (parser->beg == parser->end || *(parser->beg) != '\"')
        {
            parser->beg = end + 1;
            t2c_format_parse_error(parser, errAttrMissingQuot, NULL);
            free(name);
            smap_destroy(*attr);
            *attr = NULL;
            return PR_SYNTAX_ERROR;
        }
        
        ++(parser->beg);
        if (parser->beg == parser->end)
        {
            t2c_format_parse_error(parser, errAttrMissingQuot, NULL);
            free(name);
            smap_destroy(*attr);
            *attr = NULL;
            return PR_SYNTAX_ERROR;
        }
        
        end = mist_find_in_range_first_of(parser->beg, parser->end, "\"", 1);
        if (end == parser->end)
        {
            t2c_format_parse_error(parser, errAttrMissingQuot, NULL);
            free(name);
            smap_destroy(*attr);
            *attr = NULL;
            return PR_SYNTAX_ERROR;
        }
        
        char* val = mist_get_substring(parser->beg, end);
        if (val == NULL)
        {
            free(name);
            smap_destroy(*attr);
            *attr = NULL;
            return PR_NO_MEMORY;
        }
        
        // add the {name, value} pair to the string map
        if (!smap_add_element(*attr, name, val))
        {
            free(name);
            free(val);
            smap_destroy(*attr);
            *attr = NULL;
            return PR_NO_MEMORY;
        }
        
        parser->beg = end + 1;
        
        free(name);
        free(val);
    }
    
    // check for duplicate attributes
    const char* dup_key = smap_check_duplicate_keys(*attr);
    if (dup_key != NULL)
    {
        parser->beg = old_beg;
        t2c_format_parse_error(parser, errAttrDuplicate, dup_key);
        smap_destroy(*attr);
        *attr = NULL;
        return PR_SYNTAX_ERROR;
    }
    
    return PR_PARSE_OK;
}

void
t2c_parser_init(CT2CParser* parser, const char* source, 
                const char* beg, const char* end)
{
    assert(parser != NULL);
    
    parser->source = source;
    parser->beg = beg;
    parser->end = end;
    parser->old_end = end;
    parser->error = NULL;
    
    assert(t2c_is_parser_valid(parser));
    return;
}

void
t2c_parser_advance(CT2CParser* parser)
{
    assert(t2c_is_parser_valid(parser));
    
    parser->beg = parser->end;
    parser->end = parser->old_end;
    
    return;
}

void
t2c_parser_to_section_end(CT2CParser* parser)
{
    assert(t2c_is_parser_valid(parser));
    
    parser->end = mist_find_in_range_first_of(parser->end, parser->old_end, ">", 1);
    assert(parser->end < parser->old_end);
    ++(parser->end);
    
    return;
}

///////////////////////////////////////////////////////////////////////////
